/* .wy-nav-content { */
/*     max-width: none; */
/* } */

/* override table width restrictions */
/* @media screen and (min-width: 767px) { */
/*     .wy-table-responsive table td { */
/*         /\* !important prevents the common CSS stylesheets from */
/*        overriding this as on RTD they are loaded after this stylesheet *\/ */
/*         white-space: normal !important; */
/*     } */

/*     .wy-table-responsive { */
/*         overflow: visible !important; */
/*     } */
/* } */


@media screen and (min-width: 767px) {
    .tight-table table td {
        /* !important prevents the common CSS stylesheets from
       overriding this as on RTD they are loaded after this stylesheet */
        white-space: normal !important;
    }
    .tight-table-responsive {
        overflow: visible !important;
    }
}
